@@ -62,17 +62,19 @@ struct _BrkPool {
     BrkShot         shot;
 };
 
-static void         brk_pool_allocator_use ();
-static void         brk_pool_allocator_clear ();
+static BrkPool *    brk_pool_node_new (const BrkShot *shot, BrkEnv *env);
+static void         brk_pool_node_free (BrkPool *pool, BrkEnv *env);
 
-static BrkPool *    brk_pool_node_new (const BrkShot *shot);
-static void         brk_pool_node_free (BrkPool *pool);
-
-static void         brk_pool_free (BrkPool *pool);
+static void         brk_pool_free (BrkPool *pool, BrkEnv *env);
 static BrkPool *    brk_pool_get_node (BrkPool *pool);
-static BrkPool *    brk_pool_match (BrkPool *pool, BrkPool *node);
+static BrkPool *    brk_pool_match (BrkPool *pool, const BrkPool *node);
 static BrkPool *    brk_pool_add (BrkPool *pool, BrkPool *node);
-static BrkPool *    brk_pool_delete (BrkPool *pool, BrkPool *node);
+static BrkPool *    brk_pool_delete_node (BrkPool *pool, BrkPool *node,
+                                          BrkEnv *env);
+
+struct _BrkEnv {
+    BrkPool        *free_list;
+};
 
 /**
  * @brief Best break solution
@@ -101,33 +103,23 @@ typedef struct {
  *   PRIVATE METHODS DECLARATIONS   *
  *----------------------------------*/
 
-static BrkPool *    brk_root_pool (int pos_size);
+static BrkPool *    brk_root_pool (int pos_size, BrkEnv *env);
 static int          brk_maximal_do_impl (const thwchar_t *ws, int len,
                                          const char *brkpos_hints,
-                                         int pos[], size_t n);
+                                         int pos[], size_t n, BrkEnv *env);
 static int          brk_recover_try (const thwchar_t *ws, int len,
                                      const char *brkpos_hints,
-                                     size_t recov_words, int *last_brk_pos);
+                                     size_t recov_words, int *last_brk_pos,
+                                     BrkEnv *env);
 static int          brk_recover (const thwchar_t *wtext, int len, int pos,
-                                 const char *brkpos_hints, RecovHist *rh);
+                                 const char *brkpos_hints, RecovHist *rh,
+                                 BrkEnv *env);
 
 /*---------------------*
  *   PRIVATE GLOBALS   *
  *---------------------*/
-void
-brk_maximal_init ()
-{
-    brk_pool_allocator_use ();
-}
-
-void
-brk_maximal_quit ()
-{
-    brk_pool_allocator_clear ();
-}
-
 int
-brk_maximal_do (const thchar_t *s, int len, int pos[], size_t n)
+brk_maximal_do (const thchar_t *s, int len, int pos[], size_t n, BrkEnv *env)
 {
     char        *brkpos_hints;
     thwchar_t   *ws;
@@ -143,7 +135,7 @@ brk_maximal_do (const thchar_t *s, int len, int pos[], size_t n)
         goto err_brkpos_hints_created;
     th_tis2uni_line (s, ws, len + 1);
 
-    ret = brk_maximal_do_impl (ws, len, brkpos_hints, pos, n);
+    ret = brk_maximal_do_impl (ws, len, brkpos_hints, pos, n, env);
 
     free (ws);
     free (brkpos_hints);
@@ -159,7 +151,8 @@ err_nothing_done:
 static int
 brk_maximal_do_impl (const thwchar_t *ws, int len,
                      const char *brkpos_hints,
-                     int pos[], size_t n)
+                     int pos[], size_t n,
+                     BrkEnv *env)
 {
     BrkPool     *pool;
     BrkPool     *node;
@@ -167,7 +160,7 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
     RecovHist    recov_hist;
     int          ret;
 
-    pool = brk_root_pool (n);
+    pool = brk_root_pool (n, env);
     best_brk = best_brk_new (n);
     if (UNLIKELY (!best_brk))
         return 0;
@@ -175,7 +168,6 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
 
     while (NULL != (node = brk_pool_get_node (pool))) {
         BrkShot *shot = &node->shot;
-        BrkPool *match;
         int      is_keep_node, is_terminal, is_recovered;
         int      str_pos;
 
@@ -192,7 +184,7 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
 
                 /* try to recover from error */
                 recovered = brk_recover (ws, len, shot->str_pos + 1,
-                                         brkpos_hints, &recov_hist);
+                                         brkpos_hints, &recov_hist, env);
                 if (-1 != recovered) {
                     /* add penalty by recovered - recent break pos */
                     shot->penalty += recovered;
@@ -236,7 +228,7 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
                 !trie_state_is_single (shot->dict_state))
             {
                 /* add node to mark break position instead of current */
-                node = brk_pool_node_new (shot);
+                node = brk_pool_node_new (shot, env);
                 pool = brk_pool_add (pool, node);
                 shot = &node->shot;
             }
@@ -248,10 +240,14 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
         if (!is_keep_node || shot->str_pos == len || shot->cur_brk_pos >= n) {
             /* path is done; contest and remove */
             best_brk_contest (best_brk, shot);
-            pool = brk_pool_delete (pool, node);
+            pool = brk_pool_delete_node (pool, node, env);
         } else {
+            BrkPool *pool_tail = pool;
+            BrkPool *match;
+
             /* find matched nodes, contest and keep the best one */
-            while (NULL != (match = brk_pool_match (pool, node))) {
+            while (NULL != (match = brk_pool_match (pool_tail, node))) {
+                BrkPool *next = match->next;
                 BrkPool *del_node;
 
                 if (match->shot.penalty < node->shot.penalty ||
@@ -259,11 +255,16 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
                      match->shot.cur_brk_pos < node->shot.cur_brk_pos))
                 {
                     del_node = node;
+                    /* (match->next == node) -> also skip node */
+                    if (next == node) {
+                        next = node->next;
+                    }
                     node = match;
                 } else {
                     del_node = match;
                 }
-                pool = brk_pool_delete (pool, del_node);
+                pool = brk_pool_delete_node (pool, del_node, env);
+                pool_tail = next;
             }
         }
     }
@@ -271,7 +272,7 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
     ret = best_brk->cur_brk_pos;
     memcpy (pos, best_brk->brk_pos, ret * sizeof (pos[0]));
 
-    brk_pool_free (pool);
+    brk_pool_free (pool, env);
     best_brk_free (best_brk);
     return ret;
 }
@@ -279,18 +280,18 @@ brk_maximal_do_impl (const thwchar_t *ws, int len,
 static int
 brk_recover_try (const thwchar_t *ws, int len,
                  const char *brkpos_hints,
-                 size_t recov_words, int *last_brk_pos)
+                 size_t recov_words, int *last_brk_pos,
+                 BrkEnv *env)
 {
     BrkPool     *pool;
     BrkPool     *node;
     int          ret;
 
-    pool = brk_root_pool (recov_words);
+    pool = brk_root_pool (recov_words, env);
     ret = 0;
 
     while (NULL != (node = brk_pool_get_node (pool))) {
         BrkShot *shot = &node->shot;
-        BrkPool *match;
         int      is_keep_node, is_terminal;
 
         /* walk dictionary character-wise till a word is matched */
@@ -313,7 +314,7 @@ brk_recover_try (const thwchar_t *ws, int len,
             } while (!(is_terminal && brkpos_hints[shot->str_pos]));
 
             if (!is_keep_node) {
-                pool = brk_pool_delete (pool, node);
+                pool = brk_pool_delete_node (pool, node, env);
                 break;
             }
 
@@ -322,7 +323,7 @@ brk_recover_try (const thwchar_t *ws, int len,
                 !trie_state_is_single (shot->dict_state))
             {
                 /* add node to mark break position instead of current */
-                node = brk_pool_node_new (shot);
+                node = brk_pool_node_new (shot, env);
                 pool = brk_pool_add (pool, node);
                 shot = &node->shot;
             }
@@ -336,27 +337,32 @@ brk_recover_try (const thwchar_t *ws, int len,
                     ret = shot->cur_brk_pos;
                     *last_brk_pos = shot->brk_pos[ret - 1];
                 }
-                pool = brk_pool_delete (pool, node);
+                pool = brk_pool_delete_node (pool, node, env);
                 /* stop as soon as first solution is found */
                 if (ret == recov_words)
                     goto recov_done;
                 break;
             } else {
-                /* find matched nodes, contest and keep the best one */
-                while (NULL != (match = brk_pool_match (pool, node))) {
-                    pool = brk_pool_delete (pool, match);
+                BrkPool *pool_tail = pool;
+                BrkPool *match;
+
+                /* find matched nodes and delete them */
+                while (NULL != (match = brk_pool_match (pool_tail, node))) {
+                    BrkPool *next = match->next;
+                    pool = brk_pool_delete_node (pool, match, env);
+                    pool_tail = next;
                 }
             }
         }
     }
 
 recov_done:
-    brk_pool_free (pool);
+    brk_pool_free (pool, env);
     return ret;
 }
 
 static BrkPool *
-brk_root_pool (int pos_size)
+brk_root_pool (int pos_size, BrkEnv *env)
 {
     Trie       *dict;
     BrkPool    *pool;
@@ -374,7 +380,7 @@ brk_root_pool (int pos_size)
     root_shot.str_pos = root_shot.cur_brk_pos = 0;
     root_shot.penalty = 0;
 
-    node = brk_pool_node_new (&root_shot);
+    node = brk_pool_node_new (&root_shot, env);
     if (LIKELY (node)) {
         pool = brk_pool_add (pool, node);
     }
@@ -388,7 +394,8 @@ brk_root_pool (int pos_size)
 
 static int
 brk_recover (const thwchar_t *wtext, int len, int pos,
-             const char *brkpos_hints, RecovHist *rh)
+             const char *brkpos_hints, RecovHist *rh,
+             BrkEnv *env)
 {
     int last_brk_pos = 0;
     int n, p;
@@ -402,7 +409,7 @@ brk_recover (const thwchar_t *wtext, int len, int pos,
     for (p = pos; p < len; ++p) {
         if (brkpos_hints[p]) {
             n = brk_recover_try (wtext + p, len - p, brkpos_hints + p,
-                                 RECOVERED_WORDS, &last_brk_pos);
+                                 RECOVERED_WORDS, &last_brk_pos, env);
             if (n == RECOVERED_WORDS
                 || (n > 0 && '\0' == wtext[last_brk_pos]))
             {
@@ -456,40 +463,42 @@ brk_shot_destruct (BrkShot *shot)
         free (shot->brk_pos);
 }
 
-static BrkPool *brk_pool_free_list = NULL;
-static int      brk_pool_allocator_refcnt = 0;
-
-static void
-brk_pool_allocator_use ()
+BrkEnv *
+brk_env_new()
 {
-    ++brk_pool_allocator_refcnt;
+    BrkEnv *env = (BrkEnv *) malloc (sizeof (BrkEnv));
+    if (UNLIKELY (!env))
+        return NULL;
+
+    env->free_list = NULL;
+
+    return env;
 }
 
-static void
-brk_pool_allocator_clear ()
+void
+brk_env_free (BrkEnv *env)
 {
-    if (--brk_pool_allocator_refcnt > 0)
-        return;
-
-    while (brk_pool_free_list) {
+    while (env->free_list) {
         BrkPool *next;
 
-        next = brk_pool_free_list->next;
-        brk_shot_destruct (&brk_pool_free_list->shot);
-        free (brk_pool_free_list);
-        brk_pool_free_list = next;
+        next = env->free_list->next;
+        brk_shot_destruct (&env->free_list->shot);
+        free (env->free_list);
+        env->free_list = next;
     }
+
+    free (env);
 }
 
 static BrkPool *
-brk_pool_node_new (const BrkShot *shot)
+brk_pool_node_new (const BrkShot *shot, BrkEnv *env)
 {
     BrkPool *node;
 
-    if (brk_pool_free_list) {
+    if (env->free_list) {
         /* reuse old node if possible */
-        node = brk_pool_free_list;
-        brk_pool_free_list = brk_pool_free_list->next;
+        node = env->free_list;
+        env->free_list = env->free_list->next;
         brk_shot_reuse (&node->shot, shot);
     } else {
         node = (BrkPool *) malloc (sizeof (BrkPool));
@@ -507,21 +516,21 @@ brk_pool_node_new (const BrkShot *shot)
 }
 
 static void
-brk_pool_node_free (BrkPool *pool)
+brk_pool_node_free (BrkPool *pool, BrkEnv *env)
 {
     /* put it in free list for further reuse */
-    pool->next = brk_pool_free_list;
-    brk_pool_free_list = pool;
+    pool->next = env->free_list;
+    env->free_list = pool;
 }
 
 static void
-brk_pool_free (BrkPool *pool)
+brk_pool_free (BrkPool *pool, BrkEnv *env)
 {
     while (pool) {
         BrkPool *next;
 
         next = pool->next;
-        brk_pool_node_free (pool);
+        brk_pool_node_free (pool, env);
         pool = next;
     }
 }
@@ -553,26 +562,28 @@ brk_pool_get_node (BrkPool *pool)
 }
 
 static BrkPool *
-brk_pool_match (BrkPool *pool, BrkPool *node)
+brk_pool_match (BrkPool *pool, const BrkPool *node)
 {
     int node_cur_pos;
 
     node_cur_pos = node->shot.cur_brk_pos;
-    while (pool) {
-        if (pool != node) {
-            if (node_cur_pos == 0) {
-                if (pool->shot.cur_brk_pos == 0)
-                    break;
-            } else {
-                if (pool->shot.cur_brk_pos > 0 &&
-                    pool->shot.brk_pos[pool->shot.cur_brk_pos - 1]
-                        == node->shot.brk_pos[node_cur_pos - 1])
-                {
-                    break;
-                }
+    if (node_cur_pos == 0) {
+        while (pool) {
+            if (pool != node && pool->shot.cur_brk_pos == 0)
+                break;
+            pool = pool->next;
+        }
+    } else {
+        int node_brk_pos = node->shot.brk_pos[node_cur_pos - 1];
+        while (pool) {
+            if (pool != node &&
+                pool->shot.cur_brk_pos > 0 &&
+                pool->shot.brk_pos[pool->shot.cur_brk_pos - 1] == node_brk_pos)
+            {
+                break;
             }
+            pool = pool->next;
         }
-        pool = pool->next;
     }
     return pool;
 }
@@ -585,7 +596,7 @@ brk_pool_add (BrkPool *pool, BrkPool *node)
 }
 
 static BrkPool *
-brk_pool_delete (BrkPool *pool, BrkPool *node)
+brk_pool_delete_node (BrkPool *pool, BrkPool *node, BrkEnv *env)
 {
     if (pool == node) {
         pool = pool->next;
@@ -597,7 +608,7 @@ brk_pool_delete (BrkPool *pool, BrkPool *node)
         if (p)
             p->next = node->next;
     }
-    brk_pool_node_free (node);
+    brk_pool_node_free (node, env);
 
     return pool;
 }
